@@ -18,8 +18,18 @@
 
 open OpamTypes
 
-(** Manage pinned packages. *)
-val pin: force:bool -> pin -> unit
+(** Pin a package. Returns [Some is_same_version] if the package should be
+    reinstalled (or upgraded if [is_same_version] is false *)
+val pin: name -> pin_option -> bool option
+
+(** Let the user edit a pinned package's opam file.
+    Returns [Some is_same_version] if the package should be rebuilt.
+    raises [Not_found] if no valid opam file is available and the user didn't
+    succeed in producing one. *)
+val edit: OpamState.state -> name -> bool option
+
+(** Unpin a package. Returns true if the package should be rebuilt *)
+val unpin: name -> bool
 
 (** List the pinned packages. *)
-val list: unit -> unit
+val list: short:bool -> unit -> unit
