@@ -15,9 +15,53 @@
 (**************************************************************************)
 
 open OpamTypes
+open OpamTypesBase
 open OpamState.Types
+open OpamMisc.OP
+open OpamPackage.Set.Op
 
 let log fmt = OpamGlobals.log "REPOSITORY" fmt
+let slog = OpamGlobals.slog
+
+let update t repo =
+  let max_loop = 10 in
+  (* Recursively traverse redirection links, but stop after 10 steps or if
+     we start to cycle. *)
+  let rec loop r n =
+    if n = 0 then
+      OpamGlobals.warning "%s: Too many redirections, stopping."
+        (OpamRepositoryName.to_string repo.repo_name)
+    else (
+      OpamRepository.update r;
+      if n <> max_loop && r = repo then
+        OpamGlobals.warning "%s: Cyclic redirections, stopping."
+          (OpamRepositoryName.to_string repo.repo_name)
+      else match OpamState.redirect t r with
+        | None -> ()
+        | Some (new_repo, f) ->
+          OpamFilename.rmdir repo.repo_root;
+          OpamFile.Repo_config.write (OpamPath.Repository.config repo) new_repo;
+          let reason = match f with
+            | None   -> ""
+            | Some f -> Printf.sprintf " (%s)" (OpamFilter.to_string f) in
+          OpamGlobals.note
+            "The repository '%s' will be *%s* redirected to %s%s"
+            (OpamRepositoryName.to_string repo.repo_name)
+            ((OpamGlobals.colorise `bold) "permanently")
+            (OpamMisc.prettify_path (string_of_address new_repo.repo_address))
+            reason;
+          loop new_repo (n-1)
+    ) in
+  loop repo max_loop;
+  let repo =
+    repo
+    |> OpamPath.Repository.config
+    |> OpamFile.Repo_config.safe_read
+  in
+  OpamRepository.check_version repo;
+  let repositories =
+    OpamRepositoryName.Map.add repo.repo_name repo t.repositories in
+  { t with repositories }
 
 let print_updated_compilers updates =
 
@@ -44,7 +88,7 @@ let print_updated_compilers updates =
 
   print
     "The following compiler description has been DELETED"
-    "The following compiler descriptions are been DELETED"
+    "The following compiler descriptions have been DELETED"
     updates.deleted
 
 let filter_compiler_checksums cs =
@@ -57,8 +101,8 @@ let fix_compiler_descriptions t ~verbose =
   let global_index = OpamState.compiler_state t in
   let repo_index = OpamState.compiler_repository_state t in
   let niet = String.concat ":" in
-  log "global-index: %s" (OpamCompiler.Map.to_string niet global_index);
-  log "repo-index  : %s" (OpamCompiler.Map.to_string niet repo_index);
+  log "global-index: %a" (slog @@ OpamCompiler.Map.to_string niet) global_index;
+  log "repo-index  : %a" (slog @@ OpamCompiler.Map.to_string niet) repo_index;
 
   let updated_compilers, new_compilers =
     let updated_compilers =
@@ -70,8 +114,8 @@ let fix_compiler_descriptions t ~verbose =
             set
         ) repo_index OpamCompiler.Set.empty in
     OpamCompiler.Set.partition (OpamState.is_compiler_installed t) updated_compilers in
-  log "updated-compilers: %s" (OpamCompiler.Set.to_string updated_compilers);
-  log "new-compilers    : %s" (OpamCompiler.Set.to_string new_compilers);
+  log "updated-compilers: %a" (slog OpamCompiler.Set.to_string) updated_compilers;
+  log "new-compilers    : %a" (slog OpamCompiler.Set.to_string) new_compilers;
 
   let deleted_compilers =
     OpamCompiler.Set.fold (fun comp map ->
@@ -82,7 +126,7 @@ let fix_compiler_descriptions t ~verbose =
         else
           OpamCompiler.Set.add comp map
       ) t.compilers OpamCompiler.Set.empty in
-  log "deleted-compilers: %s" (OpamCompiler.Set.to_string deleted_compilers);
+  log "deleted-compilers: %a" (slog OpamCompiler.Set.to_string) deleted_compilers;
 
   (* Delete compiler descritions (unless they are still installed) *)
   OpamCompiler.Set.iter (fun comp ->
@@ -188,9 +232,9 @@ let print_updated_dev_packages pinned_packages =
 
 (* Check for updates in pinned packages *)
 let update_dev_packages t ~verbose packages =
-  log "update-dev-packages updates %s" (OpamPackage.Set.to_string packages);
-  let updates = OpamState.update_dev_packages t in
-  if verbose then print_updated_dev_packages updates;
+  log "update-dev-packages updates %a" (slog OpamPackage.Set.to_string) packages;
+  let updates = OpamState.update_dev_packages t packages in
+  if verbose then print_updated_dev_packages updates ;
   updates
 
 let filter_package_checksums cs =
@@ -215,11 +259,11 @@ let fix_package_descriptions t ~verbose =
   let updated_packages, new_packages =
     let updated_packages =
       OpamPackage.Map.fold (fun nv state set ->
-          if not (OpamPackage.Map.mem nv global_index)
-          || OpamPackage.Map.find nv global_index <> state then
-            OpamPackage.Set.add nv set
-          else
-            set
+          try
+            if OpamPackage.Map.find nv global_index <> state
+            then OpamPackage.Set.add nv set
+            else set
+          with Not_found -> OpamPackage.Set.add nv set
         ) repo_index OpamPackage.Set.empty in
     OpamPackage.Set.partition (fun nv ->
         OpamPackage.Map.mem nv global_index
@@ -237,17 +281,15 @@ let fix_package_descriptions t ~verbose =
           true
       ) all_installed in
 
-  log "new-packages     : %s" (OpamPackage.Set.to_string new_packages);
-  log "updated-packages : %s" (OpamPackage.Set.to_string updated_packages);
-  log "changed-packages : %s" (OpamPackage.Set.to_string changed_packages);
-  log "missing-installed: %s" (OpamPackage.Set.to_string missing_installed_packages);
+  log "new-packages     : %a" (slog OpamPackage.Set.to_string) new_packages;
+  log "updated-packages : %a" (slog OpamPackage.Set.to_string) updated_packages;
+  log "changed-packages : %a" (slog OpamPackage.Set.to_string) changed_packages;
+  log "missing-installed: %a" (slog OpamPackage.Set.to_string) missing_installed_packages;
 
   let deleted_packages =
-    OpamPackage.Set.filter (fun nv ->
-        not (OpamPackage.is_pinned nv                         (* pinned*)
-             || OpamPackage.Map.mem nv repo_index)     (* OR available *)
-      ) t.packages in
-  log "deleted-packages: %s" (OpamPackage.Set.to_string deleted_packages);
+    t.packages -- OpamPackage.keys repo_index -- OpamState.pinned_packages t
+  in
+  log "deleted-packages: %a" (slog OpamPackage.Set.to_string) deleted_packages;
 
   (* Notify only about deleted packages that are installed or were just removed
      (ie ignore the one that were removed from upstream but still have data
@@ -268,15 +310,15 @@ let fix_package_descriptions t ~verbose =
 
   (* that's not a good idea *at all* to enable this hook if you
            are not in a testing environment *)
-  OpamPackage.Map.iter (fun nv _ ->
-      if !OpamGlobals.sync_archives then (
-        log "download %s" (OpamFilename.to_string (OpamPath.archive t.root nv));
+  if !OpamGlobals.sync_archives then
+    OpamPackage.Map.iter (fun nv _ ->
+        log "download %a"
+          (slog @@ OpamFilename.to_string @* OpamPath.archive t.root) nv;
         match OpamState.download_archive t nv with
         | None | Some _ -> ()
-      )
-    ) repo_index;
+      ) repo_index;
 
-  (* Do not recompile a package if only only OPAM or descr files have
+  (* Do not recompile a package if only OPAM or descr files have
      changed. We recompile a package:
 
      - if both global and repo states have an archive file and the
@@ -301,7 +343,7 @@ let fix_package_descriptions t ~verbose =
           OpamState.package_repository_partial_state t nv ~archive:false in
         checksums_g <> checksums_r
     ) changed_packages in
-  log "packages-to-reinstall: %s" (OpamPackage.Set.to_string changed_packages);
+  log "packages-to-reinstall: %a" (slog OpamPackage.Set.to_string) changed_packages;
 
   (* Update the package descriptions *)
   OpamPackage.Set.iter (fun nv ->
@@ -322,9 +364,19 @@ let fix_package_descriptions t ~verbose =
           OpamFilename.remove (OpamPath.Repository.archive repo nv);
     ) (OpamPackage.Set.union missing_installed_packages updated_packages);
 
+  (* Remove archives of non-installed packages (these may no longer be
+     up-to-date) *)
+  OpamPackage.Set.iter (fun nv ->
+      let f = OpamPath.archive t.root nv in
+      if OpamFilename.exists f then
+        (log "Cleaning up obsolete archive %a" (slog OpamFilename.to_string) f;
+         OpamFilename.remove f))
+    (OpamPackage.keys global_index -- all_installed);
+
   (* Display some warnings/errors *)
   OpamPackage.Set.iter (fun nv ->
-      let file = OpamPath.Switch.Overlay.opam t.root t.switch nv in
+      let file =
+        OpamPath.Switch.Overlay.opam t.root t.switch (OpamPackage.name nv) in
       let file =
         if OpamFilename.exists file then file else OpamPath.opam t.root nv in
       if not (OpamFilename.exists file) then
@@ -376,8 +428,9 @@ let update_compiler_index t =
 (* update the repository config file:
    ~/.opam/repo/<repo>/config *)
 let update_config t repos =
-  log "update-config %s"
-    (OpamMisc.pretty_list (List.map OpamRepositoryName.to_string repos));
+  log "update-config %a"
+    (slog @@ OpamMisc.pretty_list @* List.map OpamRepositoryName.to_string)
+    repos;
   let new_config = OpamFile.Config.with_repositories t.config repos in
   OpamFile.Config.write (OpamPath.config t.root) new_config
 
@@ -393,7 +446,7 @@ let () =
 
 (* Remove any remaining of [repo] from OPAM state *)
 let cleanup t repo =
-   log "cleanup %s" (OpamRepositoryName.to_string repo.repo_name);
+   log "cleanup %a" (slog OpamRepositoryName.to_string) repo.repo_name;
   let repos = OpamRepositoryName.Map.keys t.repositories in
   update_config t (List.filter ((<>) repo.repo_name) repos);
   OpamFilename.rmdir repo.repo_root;
@@ -416,52 +469,54 @@ let priority repo_name ~priority =
 let add name kind address ~priority:prio =
   log "repository-add";
   let t = OpamState.load_state "repository-add" in
-  let repo = {
-    repo_name     = name;
-    repo_kind     = kind;
-    repo_address  = address;
-    repo_priority = min_int; (* we initially put it as low-priority *)
-    repo_root     = OpamPath.Repository.create t.root name;
-  } in
   if OpamState.mem_repository t name then
     OpamGlobals.error_and_exit
       "%s is already a remote repository"
-      (OpamRepositoryName.to_string name)
-  else (
-    try OpamRepository.init repo with
-    | OpamRepository.Unknown_backend ->
-      OpamGlobals.error_and_exit
-        "\"%s\" is not a supported backend"
-        (string_of_repository_kind repo.repo_kind)
-    | e ->
-      cleanup t repo;
-      raise e
-  );
+      (OpamRepositoryName.to_string name);
   if kind = `local then (
     let repo_dir = OpamFilename.Dir.of_string (string_of_address address) in
     let pkgdir = OpamPath.packages_dir repo_dir in
     let compdir = OpamPath.compilers_dir repo_dir in
-    if not (OpamFilename.exists_dir pkgdir) && not (OpamFilename.exists_dir compdir) then
-      OpamGlobals.warning "%S doesn't contain a \"packages\" nor a \"compilers\" directory.\n\
-                           Is it really the directory of your repo ? \
-                           (\"opam remote remove %s\" to revert)"
-        (OpamFilename.Dir.to_string repo_dir) (OpamRepositoryName.to_string name)
+    if not (OpamFilename.exists_dir pkgdir) &&
+       not (OpamFilename.exists_dir compdir) &&
+       not (OpamGlobals.confirm
+              "%S doesn't contain a \"packages\" nor a \"compilers\" directory.\n\
+               Is it really the directory of your repo ?"
+              (OpamFilename.Dir.to_string repo_dir))
+    then OpamGlobals.exit 1
   );
-  log "Adding %s" (OpamRepository.to_string repo);
-  update_config t (repo.repo_name :: OpamRepositoryName.Map.keys t.repositories);
+  let prio = match prio with
+    | Some p -> p
+    | None ->
+      if OpamRepositoryName.Map.is_empty t.repositories then 0 else
+      let max_prio =
+        OpamRepositoryName.Map.fold
+          (fun _ { repo_priority } m -> max repo_priority m)
+          t.repositories min_int in
+      10 + max_prio in
+  let repo = {
+    repo_name     = name;
+    repo_kind     = kind;
+    repo_address  = address;
+    repo_priority = prio;
+    repo_root     = OpamPath.Repository.create t.root name;
+  } in
+  OpamRepository.init repo;
+  log "Adding %a" (slog OpamRepository.to_string) repo;
+  let repositories = OpamRepositoryName.Map.add name repo t.repositories in
+  update_config t (OpamRepositoryName.Map.keys repositories);
+  let t = { t with repositories } in
+  OpamState.remove_state_cache ();
   try
-    let prio = match prio with
-      | Some p -> p
-      | None ->
-        if OpamRepositoryName.Map.is_empty t.repositories then 0 else
-          let max_prio =
-            OpamRepositoryName.Map.fold
-              (fun _ { repo_priority } m -> max repo_priority m)
-              t.repositories min_int in
-          10 + max_prio in
-    OpamState.remove_state_cache ();
-    priority name ~priority:prio;
-  with e ->
+    let t = update t repo in
+    fix_descriptions t ~verbose:true
+  with
+  | OpamRepository.Unknown_backend ->
+    cleanup t repo;
+    OpamGlobals.error_and_exit
+      "\"%s\" is not a supported backend"
+      (string_of_repository_kind repo.repo_kind)
+  | e ->
     cleanup t repo;
     raise e
 
@@ -471,6 +526,22 @@ let remove name =
   let repo = OpamState.find_repository t name in
   cleanup t repo
 
+let set_url name url =
+  log "repository-remove";
+  let t = OpamState.load_state ~save_cache:false "repository-set-url" in
+  let repo = OpamState.find_repository t name in
+  let url, kind = parse_url url in
+  if repo.repo_kind <> kind then
+    (remove name;
+     add name kind url ~priority:(Some repo.repo_priority))
+  else
+  (* 1/ update the config file *)
+  let config_f = OpamPath.Repository.config repo in
+  let config =
+    let config = OpamFile.Repo_config.read config_f in
+    { config with repo_address = url } in
+  OpamFile.Repo_config.write config_f config
+
 let list ~short =
   log "repository-list";
   let t = OpamState.load_state "repository-list" in
@@ -479,10 +550,7 @@ let list ~short =
       List.map
         (fun r -> OpamRepositoryName.to_string r.repo_name)
         (OpamState.sorted_repositories t) in
-    let pinned =
-      List.map OpamPackage.Name.to_string (OpamPackage.Name.Map.keys t.pinned) in
-    let all = repos @ pinned in
-    OpamGlobals.msg "%s\n" (String.concat " " all)
+    OpamGlobals.msg "%s\n" (String.concat " " repos)
   ) else (
     let pretty_print r =
       OpamGlobals.msg "%4d %-7s %10s     %s\n"
