@@ -15,10 +15,12 @@
 (**************************************************************************)
 
 open OpamTypes
+open OpamTypesBase
 open OpamMisc.OP
 open OpamFilename.OP
 
 let log fmt = OpamGlobals.log "REPOSITORY" fmt
+let slog = OpamGlobals.slog
 
 let compare r1 r2 =
   match compare r2.repo_priority r1.repo_priority with
@@ -29,7 +31,7 @@ let to_string r =
   Printf.sprintf "%s(%d %s %s)"
     (OpamRepositoryName.to_string r.repo_name)
     r.repo_priority
-    (OpamTypes.string_of_repository_kind r.repo_kind)
+    (string_of_repository_kind r.repo_kind)
     (string_of_address r.repo_address)
 
 let default_address =
@@ -98,19 +100,37 @@ let register_backend name backend =
 
 (* initialize the current directory *)
 let init repo =
-  log "init %s" (to_string repo);
+  log "init %a" (OpamGlobals.slog to_string) repo;
   let module B = (val find_backend repo: BACKEND) in
   OpamFilename.rmdir repo.repo_root;
   OpamFilename.mkdir repo.repo_root;
   OpamFile.Repo_config.write (OpamPath.Repository.config repo) repo;
   OpamFilename.mkdir (OpamPath.Repository.packages_dir repo);
   OpamFilename.mkdir (OpamPath.Repository.archives_dir repo);
-  OpamFilename.mkdir (OpamPath.Repository.compilers_dir repo);
-  ignore (B.pull_repo repo)
+  OpamFilename.mkdir (OpamPath.Repository.compilers_dir repo)
 
 let pull_url kind package local_dirname checksum remote_url =
-  let module B = (val find_backend_by_kind kind: BACKEND) in
-  B.pull_url package local_dirname checksum remote_url
+  if !OpamGlobals.req_checksums && checksum = None then
+    OpamGlobals.error_and_exit
+      "Checksum required for %s, but not found in package description.\n\
+       This may be due to an outdated package description, try running `opam update`.\n\
+       In case an update does not fix the problem, you can /bypass the check by not\n\
+       passing the `--require-checksums` command line option."
+      (OpamPackage.to_string package);
+  let pull url =
+    let module B = (val find_backend_by_kind kind: BACKEND) in
+    B.pull_url package local_dirname checksum url in
+  let rec attempt = function
+    | [] -> assert false
+    | [url] -> pull url
+    | url::mirrors ->
+      match pull url with
+      | Not_available s ->
+        OpamGlobals.warning "download of %s failed, trying mirror" s;
+        attempt mirrors
+      | r -> r
+  in
+  attempt remote_url
 
 let revision repo =
   let kind = repo.repo_kind in
@@ -149,7 +169,7 @@ let check_digest filename = function
            fixed by running `opam update`.\n\
            In case an update does not fix that problem, you can use the \
            `--no-checksums` command-line option\n\
-           to /bypass any checksum checks."
+           to /bypass checking for invalid checksums."
           (OpamFilename.to_string filename)
           expected
           actual
@@ -166,10 +186,12 @@ let check_version repo =
     |> OpamFile.Repo.safe_read
     |> OpamFile.Repo.opam_version
   end in
-  if OpamVersion.compare repo_version OpamVersion.current > 0 then
+  if not !OpamGlobals.skip_version_checks &&
+     OpamVersion.compare repo_version OpamVersion.current > 0 then
     OpamGlobals.error_and_exit
-      "\nThe current version of OPAM cannot read the repository. \
+      "The current version of OPAM cannot read the repository %S. \n\
        You should upgrade to at least version %s.\n"
+      (OpamRepositoryName.to_string repo.repo_name)
       (OpamVersion.to_string repo_version)
 
 let extract_prefix repo dir nv =
@@ -281,10 +303,9 @@ let compiler_index repositories =
     ) OpamCompiler.Map.empty repositories
 
 let update repo =
-  log "update %s" (to_string repo);
+  log "update %a" (slog to_string) repo;
   let module B = (val find_backend repo: BACKEND) in
-  B.pull_repo repo;
-  check_version repo
+  B.pull_repo repo
 
 let make_archive ?(gener_digest=false) repo prefix nv =
   let url_file = OpamPath.Repository.url repo prefix nv in
@@ -300,19 +321,21 @@ let make_archive ?(gener_digest=false) repo prefix nv =
       let url = OpamFile.URL.read url_file in
       let checksum = OpamFile.URL.checksum url in
       let remote_url = OpamFile.URL.url url in
-      let kind = guess_repository_kind (OpamFile.URL.kind url) remote_url in
-      log "downloading %s:%s"
-        (string_of_address remote_url) (string_of_repository_kind kind);
+      let mirrors = remote_url :: OpamFile.URL.mirrors url in
+      let kind = OpamFile.URL.kind url in
+      log "downloading %a:%a"
+        (slog string_of_address) remote_url
+        (slog string_of_repository_kind) kind;
       if not (OpamFilename.exists_dir download_dir) then
         OpamFilename.mkdir download_dir;
       OpamFilename.in_dir download_dir (fun () ->
           match checksum with
-          | None   -> Some (pull_url kind nv download_dir None remote_url)
+          | None   -> Some (pull_url kind nv download_dir None mirrors)
           | Some c ->
             if gener_digest then
-              Some (pull_url_and_fix_digest kind nv download_dir c url_file remote_url)
+              Some (pull_url_and_fix_digest kind nv download_dir c url_file mirrors)
             else
-              Some (pull_url kind nv download_dir checksum remote_url)
+              Some (pull_url kind nv download_dir checksum mirrors)
         )
     ) else
       None
