@@ -87,6 +87,9 @@ val init: repository -> unit
 (** Update {i $opam/repo/$repo}. *)
 val update: repository -> unit
 
+(** Error and exit on incompatible version *)
+val check_version: repository -> unit
+
 (** Backend signature *)
 module type BACKEND = sig
 
@@ -108,13 +111,14 @@ module type BACKEND = sig
 
 end
 
-(** Download an url *)
+(** Download an url. Several mirrors can be provided, in which case they will be
+    tried in order in case of an error. *)
 val pull_url: repository_kind ->
-  package -> dirname -> string option -> address -> generic_file download
+  package -> dirname -> string option -> address list -> generic_file download
 
 (** Pull and fix the resulting digest *)
 val pull_url_and_fix_digest: repository_kind ->
-  package -> dirname -> string -> filename -> address -> generic_file download
+  package -> dirname -> string -> filename -> address list -> generic_file download
 
 (** [check_digest file expected] check that the [file] digest is the
     one [expected]. *)
