@@ -35,11 +35,17 @@ let command_not_found cmd =
 module Sys2 = struct
   (* same as [Sys.is_directory] except for symlinks, which returns always [false]. *)
   let is_directory file =
-    Unix.( (lstat file).st_kind = S_DIR )
+    try Unix.( (lstat file).st_kind = S_DIR )
+    with Unix.Unix_error _ as e -> raise (Sys_error (Printexc.to_string e))
 end
 
 let (/) = Filename.concat
 
+(* Separator for the PATH variables *)
+let path_sep = match OpamGlobals.os () with
+  | OpamGlobals.Win32 -> ';'
+  | OpamGlobals.Cygwin | _ -> ':'
+
 let temp_basename prefix =
   Printf.sprintf "%s-%d-%06x" prefix (Unix.getpid ()) (Random.int 0xFFFFFF)
 
@@ -66,11 +72,26 @@ let mkdir dir =
     end in
   aux dir
 
+(* XXX: won't work on windows *)
+let remove_dir dir =
+  log "rmdir %s" dir;
+  if Sys.file_exists dir then (
+    let err = Sys.command (Printf.sprintf "rm -rf %s" dir) in
+      if err <> 0 then
+        internal_error "Cannot remove %s (error %d)." dir err
+  )
+
 let temp_files = Hashtbl.create 1024
+let check_remove_temp_dir = ref true
 
 let rec temp_file ?dir prefix =
   let temp_dir = match dir with
-    | None   -> !OpamGlobals.root_dir / "log"
+    | None   ->
+      let dir = !OpamGlobals.root_dir in
+      if !check_remove_temp_dir && dir = OpamGlobals.root_dir_tmp then (
+        OpamMisc.at_exit (fun () -> remove_dir OpamGlobals.root_dir_tmp);
+        check_remove_temp_dir := false);
+      dir / "log"
     | Some d -> d in
   mkdir temp_dir;
   let file = temp_dir / temp_basename prefix in
@@ -83,12 +104,12 @@ let rec temp_file ?dir prefix =
 
 let remove_file file =
   if Sys.file_exists file
-  || (try let _ = Unix.lstat file in true with _ -> false)
+  || (try let _ = Unix.lstat file in true with Unix.Unix_error _ -> false)
   then (
     try
       log "rm %s" file;
       Unix.unlink file
-    with e ->
+    with Unix.Unix_error _ as e ->
       internal_error "Cannot remove %s (%s)." file (Printexc.to_string e)
   )
 
@@ -129,11 +150,11 @@ let in_dir dir fn =
   let reset_cwd =
     let cwd =
       try Some (Sys.getcwd ())
-      with _ -> None in
+      with Sys_error _ -> None in
     fun () ->
       match cwd with
       | None     -> ()
-      | Some cwd -> try chdir cwd with _ -> () in
+      | Some cwd -> try chdir cwd with Unix.Unix_error _ -> () in
   chdir dir;
   try
     let r = fn () in
@@ -155,16 +176,16 @@ let list kind dir =
     []
 
 let files_with_links =
-  list (fun f -> try not (Sys.is_directory f) with _ -> true)
+  list (fun f -> try not (Sys.is_directory f) with Sys_error _ -> true)
 
 let files_all_not_dir =
-  list (fun f -> try not (Sys2.is_directory f) with _ -> true)
+  list (fun f -> try not (Sys2.is_directory f) with Sys_error _ -> true)
 
 let directories_strict =
-  list (fun f -> try Sys2.is_directory f with _ -> false)
+  list (fun f -> try Sys2.is_directory f with Sys_error _ -> false)
 
 let directories_with_links =
-  list (fun f -> try Sys.is_directory f with _ -> false)
+  list (fun f -> try Sys.is_directory f with Sys_error _ -> false)
 
 let rec_files dir =
   let rec aux accu dir =
@@ -185,14 +206,9 @@ let rec_dirs dir =
 let dirs dir =
   directories_with_links dir
 
-(* XXX: won't work on windows *)
-let remove_dir dir =
-  log "rmdir %s" dir;
-  if Sys.file_exists dir then (
-    let err = Sys.command (Printf.sprintf "rm -rf %s" dir) in
-      if err <> 0 then
-        internal_error "Cannot remove %s (error %d)." dir err
-  )
+let dir_is_empty dir =
+  Sys.file_exists dir &&
+  in_dir dir (fun () -> Sys.readdir (Sys.getcwd ()) = [||])
 
 let with_tmp_dir fn =
   let dir = mk_temp_dir () in
@@ -244,10 +260,12 @@ let default_env =
 let reset_env = lazy (
   let env = OpamMisc.env () in
   let env =
+    let path_sep_str = String.make 1 path_sep in
     List.rev_map (fun (k,v as c) ->
       match k with
       | "PATH" ->
-        k, String.concat ":" (OpamMisc.reset_env_value ~prefix:!OpamGlobals.root_dir v)
+        k, String.concat path_sep_str
+          (OpamMisc.reset_env_value path_sep ~prefix:!OpamGlobals.root_dir v)
       | _      -> c
     ) env in
   let env = List.rev_map (fun (k,v) -> k^"="^v) env in
@@ -255,11 +273,27 @@ let reset_env = lazy (
 )
 
 let command_exists ?(env=default_env) name =
-  let open OpamGlobals in
-  let cmd, args = "/bin/sh", ["-c"; Printf.sprintf "type %s" name] in
-  let r = OpamProcess.run ~env ~name:(temp_file "type") ~verbose:false cmd args in
+  let cmd, args = "/bin/sh", ["-c"; Printf.sprintf "command -v %s" name] in
+  let r = OpamProcess.run ~env ~name:(temp_file "command") ~verbose:false cmd args in
   OpamProcess.clean_files r;
-  OpamProcess.is_success r
+  if OpamProcess.is_success r then
+    let is_external_cmd s = String.contains s '/' in
+    match r.OpamProcess.r_stdout with 
+      cmdname::_ -> (* check that we have permission to execute the command *)
+	if is_external_cmd cmdname then 
+	  (try 
+	    let open Unix in 
+	    let uid = getuid() and groups = Array.to_list(getgroups()) in
+	    let s = stat cmdname in
+	    let cmd_uid = s.st_uid and cmd_gid = s.st_gid and cmd_perms = s.st_perm in
+            let mask = 0o001
+		lor (if uid = cmd_uid then 0o100 else 0)
+		lor (if List.mem cmd_gid groups then 0o010 else 0) in
+	    (cmd_perms land mask) <> 0
+          with _ -> false)
+	else true
+    | _ -> false
+  else false
 
 let runs = ref []
 let print_stats () =
@@ -276,7 +310,7 @@ let log_file name = match name with
 let log_cleanup r =
   if not !OpamGlobals.debug then OpamProcess.clean_files r
 
-let run_process ?verbose ?(env=default_env) ~name ?metadata command =
+let run_process ?verbose ?(env=default_env) ~name ?metadata ?allow_stdin command =
   let chrono = OpamGlobals.timer () in
   runs := command :: !runs;
   match command with
@@ -293,17 +327,19 @@ let run_process ?verbose ?(env=default_env) ~name ?metadata command =
         | None   -> !OpamGlobals.debug || !OpamGlobals.verbose
         | Some b -> b in
 
-      let r = OpamProcess.run ~env ~name ~verbose ?metadata cmd args in
+      let r = OpamProcess.run ~env ~name ~verbose ?metadata ?allow_stdin cmd args in
       let str = String.concat " " (cmd :: args) in
-      log "[%s] (in %.3fs) %s" (Filename.basename name) (chrono ()) str;
+      log "[%a] (in %.3fs) %s"
+        (OpamGlobals.slog Filename.basename) name
+        (chrono ()) str;
       r
     ) else
       (* Display a user-friendly message if the command does not exist *)
       command_not_found cmd
 
-let command ?verbose ?env ?name ?metadata cmd =
+let command ?verbose ?env ?name ?metadata ?allow_stdin cmd =
   let name = log_file name in
-  let r = run_process ?verbose ?env ~name ?metadata cmd in
+  let r = run_process ?verbose ?env ~name ?metadata ?allow_stdin cmd in
   if OpamProcess.is_success r then log_cleanup r
   else process_error r
 
@@ -311,21 +347,25 @@ let commands ?verbose ?env ?name ?metadata ?(keep_going=false) commands =
   let name = log_file name in
   let run = run_process ?verbose ?env ~name ?metadata in
   let command r0 c =
-    if keep_going || OpamProcess.is_success r0 then
-      let r = run c in
-      (if OpamProcess.is_failure r0 then r0 else r)
-    else r0
+    match r0, keep_going with
+    | (`Error _ | `Exception _), false -> r0
+    | _ ->
+      let r1 = try
+          let r = run c in
+          if OpamProcess.is_success r then `Successful r else `Error r
+        with Command_not_found _ as e -> `Exception e
+      in
+      match r0 with `Start | `Successful _ -> r1 | _ -> r0
   in
-  match commands with
-  | [] -> ()
-  | c1::c ->
-    let r = List.fold_left command (run c1) c in
-    if OpamProcess.is_success r then log_cleanup r
-    else process_error r
+  match List.fold_left command `Start commands with
+  | `Start -> ()
+  | `Successful r -> log_cleanup r
+  | `Error e -> process_error e
+  | `Exception e -> raise e
 
-let read_command_output ?verbose ?env ?metadata cmd =
+let read_command_output ?verbose ?env ?metadata ?allow_stdin cmd =
   let name = log_file None in
-  let r = run_process ?verbose ?env ~name ?metadata cmd in
+  let r = run_process ?verbose ?env ~name ?metadata ?allow_stdin cmd in
   if OpamProcess.is_success r then
     (log_cleanup r; r.OpamProcess.r_stdout)
   else
@@ -344,8 +384,25 @@ let copy src dst =
   if Sys.file_exists dst then
     remove_file dst;
   mkdir (Filename.dirname dst);
-  if src <> dst then
-    command ["cp"; src; dst ]
+  command ["cp"; src; dst ]
+
+let is_exec file =
+  let stat = Unix.stat file in
+  stat.Unix.st_kind = Unix.S_REG &&
+  stat.Unix.st_perm land 0o111 <> 0
+
+let install ?exec src dst =
+  if Sys.is_directory src then
+    internal_error "Cannot install %s: it is a directory." src;
+  if Sys.file_exists dst && Sys.is_directory dst then
+    internal_error "Cannot install to %s: it is a directory." dst;
+  mkdir (Filename.dirname dst);
+  let exec = match exec with
+    | Some e -> e
+    | None -> is_exec src in
+  command
+    ("install" :: "-m" :: (if exec then "0755" else "0644") ::
+     [ src; dst ])
 
 module Tar = struct
 
@@ -450,52 +507,41 @@ let link src dst =
   ) else
     internal_error "link: %s does not exist." src
 
-let flock file =
-  let l = ref 0 in
-  let id = string_of_int (Unix.getpid ()) in
-  let max_l = 5 in
-  let rec loop () =
-    if Sys.file_exists file && !l < max_l then begin
-      let ic = open_in file in
-      let pid = input_line ic in
-      close_in ic;
-      OpamGlobals.msg
-        "Another process (%s) has already locked %S. Retrying in 1s (%d/%d)\n"
-        pid file !l max_l;
-      Unix.sleep 1;
-      incr l;
-      loop ()
-    end else if Sys.file_exists file then
-      internal_error "Too many attempts. Cancelling."
-    else begin
-      let oc = open_out file in
-      output_string oc id;
-      flush oc;
-      close_out oc;
-      OpamGlobals.log id "locking %s" file;
-    end in
-  loop ()
+type lock = Unix.file_descr * string
 
-let funlock file =
-  let id = string_of_int (Unix.getpid ()) in
-  if Sys.file_exists file then (
-    let ic = open_in file in
-    try
-      let s = input_line ic in
-      close_in ic;
-      if s = id then (
-        log "unlocking %s (%s)" file id;
-        log "rm %s" file;
-        Unix.unlink file;
-      ) else
-        internal_error "Cannot unlock %s (%s)." file s
-    with _ ->
-      OpamGlobals.error "%s is broken, removing it and continuing anyway." file;
-      close_in ic;
-      log "rm %s" file;
-      try Unix.unlink file with _ -> ()
-  ) else
-    log "Cannot find %s, but continuing anyway..." file
+let flock ?(read=false) file =
+  let max_tries = if !OpamGlobals.safe_mode then 1 else 5 in
+  if not read && !OpamGlobals.safe_mode then
+    OpamGlobals.error_and_exit "Write lock attempt in safe mode";
+  let fd =
+    Unix.openfile file [Unix.O_CREAT; Unix.O_RDWR] 0o600 in
+  let lock_op = if read then Unix.F_TRLOCK else Unix.F_TLOCK in
+  let rec loop attempt =
+    try Unix.lockf fd lock_op 0
+    with Unix.Unix_error (Unix.EAGAIN,_,_) ->
+      if attempt > max_tries then
+        OpamGlobals.error_and_exit
+          "Timeout trying to acquire %s lock to %S, \
+           is another opam process running ?"
+          (if read then "read" else "write") file;
+      log
+        "Failed to %s-lock %S. (attempt %d/%d)"
+        (if read then "read" else "write")
+        file attempt max_tries;
+      Unix.sleep 1;
+      loop (attempt + 1)
+  in
+  log "locking %s" file;
+  loop 1;
+  fd, file
+
+let funlock (fd,file) =
+  (try (* Unlink file if write lock can be acquired *)
+     Unix.lockf fd Unix.F_TLOCK 0;
+     Unix.unlink file;
+   with Unix.Unix_error _ -> ());
+  Unix.close fd; (* implies Unix.lockf fd Unix.F_ULOCK 0 *)
+  log "Lock released on %s" file
 
 let ocaml_version = lazy (
   match read_command_output_opt ~verbose:false [ "ocamlc" ; "-version" ] with
@@ -507,6 +553,32 @@ let ocaml_version = lazy (
   | None    -> None
 )
 
+let exists_alongside_ocamlc name =
+  let path = try OpamMisc.getenv "PATH" with Not_found -> "" in
+  let path = OpamMisc.split path path_sep in
+  let ocamlc_dir =
+    List.fold_left (function
+        | None -> fun d ->
+          if Sys.file_exists (d/"ocamlc") then Some d else None
+        | s -> fun _ -> s)
+      None path
+  in
+  match ocamlc_dir with
+  | Some d -> Sys.file_exists (d / name)
+  | None -> false
+
+let ocaml_opt_available = lazy (exists_alongside_ocamlc "ocamlc.opt")
+let ocaml_native_available = lazy (exists_alongside_ocamlc "ocamlopt")
+let ocaml_natdynlink_available = lazy (
+  match
+    read_command_output_opt ~verbose:false [ "ocamlc"; "-where" ]
+  with
+  | Some (h::_) ->
+    let libdir = OpamMisc.strip h in
+    Sys.file_exists (libdir / "dynlink.cmxa")
+  | None | Some []   -> false
+)
+
 (* Reset the path to get the system compiler *)
 let system command = lazy (
   let env = Lazy.force reset_env in
@@ -522,7 +594,7 @@ let system_ocamlc_version = system [ "ocamlc"; "-version" ]
 
 let download_command =
   let retry = string_of_int OpamGlobals.download_retry in
-  let wget src =
+  let wget ~compress:_ src =
     let wget = [
       "wget";
       "--content-disposition"; "--no-check-certificate";
@@ -530,19 +602,22 @@ let download_command =
       src
     ] in
     command wget in
-  let curl command src =
+  let curl command ~compress src =
     let curl = [
       command;
       "--write-out"; "%{http_code}\\n"; "--insecure";
       "--retry"; retry; "--retry-delay"; "2";
-      "-OL"; src
+    ] @ (if compress then ["--compressed"] else []) @ [
+        "-OL"; src
     ] in
     match read_command_output curl with
     | [] -> internal_error "curl: empty response while downloading %s" src
     | l  ->
       let code = List.hd (List.rev l) in
       try if int_of_string code >= 400 then raise Exit
-      with _ -> internal_error "curl: code %s while downloading %s" code src in
+      with e ->
+        OpamMisc.fatal e;
+        internal_error "curl: code %s while downloading %s" code src in
   lazy (
     match OpamGlobals.curl_command with
     | Some cmd -> curl cmd
@@ -555,10 +630,10 @@ let download_command =
         internal_error "Cannot find curl nor wget."
   )
 
-let really_download ~overwrite ~src ~dst =
+let really_download ~overwrite ?(compress=false) ~src ~dst =
   let download = (Lazy.force download_command) in
   let aux () =
-    download src;
+    download ~compress src;
     match list (fun _ -> true) "." with
       ( [] | _::_::_ ) ->
       internal_error "Too many downloaded files."
@@ -574,9 +649,11 @@ let really_download ~overwrite ~src ~dst =
   try with_tmp_dir (fun tmp_dir -> in_dir tmp_dir aux)
   with
   | Internal_error s as e -> OpamGlobals.error "%s" s; raise e
-  | _ -> internal_error "Cannot download %s, please check your connection settings." src
+  | e ->
+    OpamMisc.fatal e;
+    internal_error "Cannot download %s, please check your connection settings." src
 
-let download ~overwrite ~filename:src ~dst:dst =
+let download ~overwrite ?compress ~filename:src ~dst:dst =
   if dst = src then
     dst
   else if Sys.file_exists src then (
@@ -588,7 +665,7 @@ let download ~overwrite ~filename:src ~dst:dst =
     ];
     dst
   ) else
-    really_download ~overwrite ~src ~dst
+    really_download ~overwrite ?compress ~src ~dst
 
 let patch p =
   let max_trying = 5 in
@@ -608,7 +685,9 @@ let patch p =
   let rec aux n =
     if n = max_trying then
       internal_error "Patch %s does not apply." p
-    else if None = try Some (patch ~dryrun:true n) with _ -> None then
+    else if None =
+            try Some (patch ~dryrun:true n)
+            with e -> OpamMisc.fatal e; None then
       aux (succ n)
     else
       patch ~dryrun:false n in
