@@ -19,6 +19,8 @@
 (** Exception raised when subprocess fails *)
 exception Process_error of OpamProcess.result
 
+exception Command_not_found of string
+
 (** raise [Process_error] *)
 val process_error: OpamProcess.result -> 'a
 
@@ -37,6 +39,15 @@ val with_tmp_dir: (string -> 'a) -> 'a
     if it is a link. *)
 val copy: string -> string -> unit
 
+(** [install ?exec src dst] copies file [src] as file [dst] using [install].
+    If [exec], make the resulting file executable (otherwise, look at the
+    permissions of the original file to decide). *)
+val install: ?exec:bool -> string -> string -> unit
+
+(** Checks if a file is an executable (regular file with execution
+    permission) *)
+val is_exec: string -> bool
+
 (** [link src dst] links [src] to [dst]. Remove [dst] if it is a file,
     not a directory. *)
 val link: string -> string -> unit
@@ -91,10 +102,22 @@ val rec_dirs: string -> string list
 (** Return the list of directories in the current directory. *)
 val dirs: string -> string list
 
+val dir_is_empty: string -> bool
+
 (** Return the version of the current OCaml compiler. If no OCaml
     compiler is present in the path, then it returns [None]. *)
 val ocaml_version: string option Lazy.t
 
+(** Returns true if the "ocamlopt" is available in the current switch *)
+val ocaml_native_available: bool Lazy.t
+
+(** Returns true if the ".opt" version of the current OCaml compiler is
+    available*)
+val ocaml_opt_available: bool Lazy.t
+
+(** Checks if native dynlink is available with the current OCaml compiler *)
+val ocaml_natdynlink_available: bool Lazy.t
+
 (** Return the path where the system ocamlc library is installed *)
 val system_ocamlc_where: string option Lazy.t
 
@@ -108,13 +131,14 @@ val directories_with_links: string -> string list
 (** a command is a list of words *)
 type command = string list
 
-(** Test wether a command exists in the environment. *)
+(** Test whether a command exists in the environment. *)
 val command_exists: ?env:string array -> string -> bool
 
 (** [command cmd] executes the command [cmd] in the correct OPAM
     environment. *)
 val command: ?verbose:bool -> ?env:string array -> ?name:string ->
-  ?metadata:(string * string) list -> command -> unit
+  ?metadata:(string * string) list -> ?allow_stdin:bool ->
+  command -> unit
 
 (** [commands cmds] executes the commands [cmds] in the correct OPAM
     environment. It stops whenever one command fails unless [keep_going] is set
@@ -127,7 +151,8 @@ val commands: ?verbose:bool -> ?env:string array -> ?name:string ->
     exists normally. If the command does not exist or if the command exited
     with a non-empty exit-code, throw an error. *)
 val read_command_output: ?verbose:bool -> ?env:string array ->
-  ?metadata:(string * string) list -> command -> string list
+  ?metadata:(string * string) list ->  ?allow_stdin:bool ->
+  command -> string list
 
 (** Test whether the file is an archive, by looking as its extension *)
 val is_tar_archive: string -> bool
@@ -147,19 +172,20 @@ val mkdir: string -> unit
 
 (** {2 File locking function} *)
 
-(** [flock ()] takes the global file lock. If the lock is already
-    taken, sleep for 1s and then retry. Abort after 5 tentatives. *)
-val flock: string -> unit
+type lock
 
-(** [funlock ()] unlocks the global file lock. Work only if the
-    current processus is the same as the one who took the lock at the
-    first place. *)
-val funlock: string -> unit
+(** Acquires a lock on the given file. Retries 5 times max.
+    By default, this is a write lock. *)
+val flock: ?read:bool -> string -> lock
+
+(** Releases an acquired locl *)
+val funlock: lock -> unit
 
 (** {2 Misc} *)
 
 (** download compiler sources *)
-val download: overwrite:bool -> filename:string -> dst:string -> string
+val download: overwrite:bool -> ?compress:bool ->
+  filename:string -> dst:string -> string
 
 (** Apply a patch file in the current directory. *)
 val patch: string -> unit
@@ -169,3 +195,6 @@ val temp_file: ?dir:string -> string -> string
 
 (** Print stats *)
 val print_stats: unit -> unit
+
+(** The separator character for the PATH variable *)
+val path_sep: char
